import axios from 'axios'
import { PageNumberPaginationParameters, unique } from '.'
import { AllowedMetaData, MetaData, MetaDataCreate, TypedMetaData, PageNumberPagination, UUID } from '@/types'

export interface AllowedMetaDataParameters extends PageNumberPaginationParameters {
  /**
   * ID of the corpus to retrieve allowed metadata from.
   */
  id: UUID
}

interface AllowedMetaDataCreate extends Omit<AllowedMetaData, 'id'> {
  corpusId: UUID
}

interface AllowedMetaDataEdit extends AllowedMetaData {
  corpusId: UUID
}

// List metadata type-name couples allowed for manual annotation
export const listCorpusAllowedMetadata = unique(
  async ({ id, ...params }: AllowedMetaDataParameters): Promise<PageNumberPagination<AllowedMetaData>> =>
    (await axios.get(`/corpus/${id}/allowed-metadata/`, { params })).data
)

// Create a new metadata type-name couple allowed in a corpus.
export const createAllowedMetadata = unique(
  async ({ corpusId, ...data }: AllowedMetaDataCreate): Promise<AllowedMetaData> =>
    (await axios.post(`/corpus/${corpusId}/allowed-metadata/`, data)).data
)

// Delete a metadata type-name couple from the list of metadata allowed in a corpus.
export const deleteAllowedMetadata = unique(
  async (corpusId: UUID, mdId: UUID) => await axios.delete(`/corpus/${corpusId}/allowed-metadata/${mdId}/`)
)

// Update an existing metadata type-name couple allowed in a corpus.
export const updateAllowedMetadata = unique(
  async ({ corpusId, id, ...data }: AllowedMetaDataEdit): Promise<AllowedMetaData> =>
    (await axios.patch(`/corpus/${corpusId}/allowed-metadata/${id}/`, data)).data
)

// Create a metadata.
export const createMetadata = unique(
  async (elementId: UUID, data: MetaDataCreate): Promise<MetaData> =>
    (await axios.post(`/element/${elementId}/metadata/`, data)).data
)

// Update a metadata.
export const updateMetadata = unique(
  async ({ id, ...data }: TypedMetaData): Promise<MetaData> =>
    (await axios.patch(`/metadata/${id}/`, data)).data
)

// Delete a metadata.
export const deleteMetadata = unique(async (id: UUID) => await axios.delete(`/metadata/${id}/`))

// List element metadata.
export const listMetadata = unique(
  async (id: UUID): Promise<MetaData[]> => (await axios.get(`/element/${id}/metadata/`)).data
)
