import { assert } from 'chai'
import axios from 'axios'
import { initialState, mutations } from '@/store/classification.js'
import store from './index.spec.js'
import { FakeAxios, assertRejects } from '../testhelpers.js'

describe('classification', () => {
  describe('mutations', () => {
    describe('setHasMLClasses', () => {
      it('sets hasMLClasses when it does not exist', () => {
        const state = {
          hasMLClasses: {
            corpus1: true
          }
        }
        mutations.setHasMLClasses(state, { id: 'corpus2', value: false })
        assert.deepStrictEqual(state, {
          hasMLClasses: {
            corpus1: true,
            corpus2: false
          }
        })
      })

      it('ignores when hasMLClasses is already set', () => {
        const state = {
          hasMLClasses: {
            corpus1: false
          }
        }
        mutations.setHasMLClasses(state, { id: 'corpus1', value: true })
        assert.deepStrictEqual(state, {
          hasMLClasses: {
            corpus1: false
          }
        })
      })
    })

    it('setCount', () => {
      const state = {
        count: {}
      }
      mutations.setCount(state, { id: 'corpus1', value: 42 })
      assert.deepStrictEqual(state, {
        count: {
          corpus1: 42
        }
      })
    })

    it('addMLClass', () => {
      const mlClass = { id: 'classId', name: 'class name' }
      const state = initialState()
      mutations.addMLClass(state, mlClass)
      assert.deepStrictEqual(state.classes, { classId: mlClass })
    })

    it('removeMLClass', () => {
      const state = {
        hasMLClasses: {
          corpus1: true
        },
        classes: {
          classId: { id: 'classId', name: 'class name' }
        }
      }
      mutations.removeMLClass(state, 'classId')
      assert.deepStrictEqual(state.classes, {})
    })

    it('reset', () => {
      const state = {
        hasMLClasses: {
          corpus1: true
        }
      }
      mutations.reset(state)
      assert.deepStrictEqual(state, {
        hasMLClasses: {},
        classes: {},
        count: {}
      })
    })
  })

  describe('actions', () => {
    let mock

    before('Setting up mocks', () => {
      mock = new FakeAxios(axios)
    })

    afterEach(() => {
      // Remove any handlers, but leave mocking in place
      mock.reset()
      store.reset()
    })

    after('Removing Axios mock', () => {
      // Remove mocking entirely
      mock.restore()
    })

    describe('listCorpusMLClasses', () => {
      it('lists ML classes', async () => {
        const reply = {
          count: 42,
          previous: null,
          next: null,
          results: [
            {
              id: 'id',
              name: 'class name'
            }
          ]
        }
        mock.onGet('/corpus/corpus1/classes/').reply(200, reply)

        const classes = await store.dispatch('classification/listCorpusMLClasses', { id: 'corpus1' })

        assert.deepStrictEqual(classes, reply)

        assert.deepStrictEqual(store.state.classification.hasMLClasses, { corpus1: true })
        assert.deepStrictEqual(store.history, [
          {
            action: 'classification/listCorpusMLClasses',
            payload: {
              id: 'corpus1'
            }
          },
          {
            mutation: 'classification/setHasMLClasses',
            payload: {
              id: 'corpus1',
              value: true
            }
          },
          {
            mutation: 'classification/setCount',
            payload: {
              id: 'corpus1',
              value: 42
            }
          },
          {
            mutation: 'classification/addMLClass',
            payload: {
              id: 'id',
              name: 'class name',
              corpus_id: 'corpus1'
            }
          }
        ])
      })

      it('does not set hasMLClasses when searching', async () => {
        const reply = {
          count: 42,
          previous: null,
          next: null,
          results: [
            {
              id: 'id',
              name: 'class name'
            }
          ]
        }
        mock.onGet('/corpus/corpus1/classes/').reply(200, reply)

        const classes = await store.dispatch('classification/listCorpusMLClasses', { id: 'corpus1', search: 'name' })

        assert.deepStrictEqual(classes, reply)

        assert.deepStrictEqual(store.state.classification.hasMLClasses, {})
        assert.deepStrictEqual(store.history, [
          {
            action: 'classification/listCorpusMLClasses',
            payload: {
              id: 'corpus1',
              search: 'name'
            }
          },
          {
            mutation: 'classification/addMLClass',
            payload: {
              id: 'id',
              name: 'class name',
              corpus_id: 'corpus1'
            }
          }
        ])
      })

      it('handles errors', async () => {
        mock.onGet('/corpus/corpus1/classes/').reply(500)

        const err = await assertRejects(async () => await store.dispatch('classification/listCorpusMLClasses', { id: 'corpus1' }))
        assert.strictEqual(err.toString(), 'Error: Request failed with status code 500')

        assert.deepStrictEqual(store.history, [
          {
            action: 'classification/listCorpusMLClasses',
            payload: {
              id: 'corpus1'
            }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 500'
            }
          }
        ])
      })
    })

    describe('listAllCorpusMLClasses', () => {
      it('paginates through listCorpusMLClasses', async () => {
        mock.onGet('/corpus/corpus1/classes/', { params: { page: 1 } }).reply(200, {
          count: 3,
          previous: null,
          next: 'blah blah',
          results: [
            {
              id: 'class1',
              name: 'class the first'
            }
          ]
        })
        mock.onGet('/corpus/corpus1/classes/', { params: { page: 2 } }).reply(200, {
          count: 3,
          previous: 'blah blah',
          next: 'blah blah',
          results: [
            {
              id: 'class2',
              name: 'class the second'
            }
          ]
        })
        mock.onGet('/corpus/corpus1/classes/', { params: { page: 3 } }).reply(200, {
          count: 3,
          previous: 'blah blah',
          next: null,
          results: [
            {
              id: 'class3',
              name: 'class the third'
            }
          ]
        })

        await store.dispatch('classification/listAllCorpusMLClasses', 'corpus1')

        assert.deepStrictEqual(store.history, [
          {
            action: 'classification/listAllCorpusMLClasses',
            payload: 'corpus1'
          },
          {
            action: 'classification/listCorpusMLClasses',
            payload: {
              id: 'corpus1',
              page: 1
            }
          },
          {
            mutation: 'classification/setHasMLClasses',
            payload: {
              id: 'corpus1',
              value: true
            }
          },
          {
            mutation: 'classification/setCount',
            payload: {
              id: 'corpus1',
              value: 3
            }
          },
          {
            mutation: 'classification/addMLClass',
            payload: {
              corpus_id: 'corpus1',
              id: 'class1',
              name: 'class the first'
            }
          },
          {
            action: 'classification/listCorpusMLClasses',
            payload: {
              id: 'corpus1',
              page: 2
            }
          },
          {
            mutation: 'classification/setHasMLClasses',
            payload: {
              id: 'corpus1',
              value: true
            }
          },
          {
            mutation: 'classification/setCount',
            payload: {
              id: 'corpus1',
              value: 3
            }
          },
          {
            mutation: 'classification/addMLClass',
            payload: {
              corpus_id: 'corpus1',
              id: 'class2',
              name: 'class the second'
            }
          },
          {
            action: 'classification/listCorpusMLClasses',
            payload: {
              id: 'corpus1',
              page: 3
            }
          },
          {
            mutation: 'classification/setHasMLClasses',
            payload: {
              id: 'corpus1',
              value: true
            }
          },
          {
            mutation: 'classification/setCount',
            payload: {
              id: 'corpus1',
              value: 3
            }
          },
          {
            mutation: 'classification/addMLClass',
            payload: {
              corpus_id: 'corpus1',
              id: 'class3',
              name: 'class the third'
            }
          }
        ])

        assert.deepStrictEqual(store.state.classification, {
          count: {
            corpus1: 3
          },
          hasMLClasses: {
            corpus1: true
          },
          classes: {
            class1: {
              id: 'class1',
              name: 'class the first',
              corpus_id: 'corpus1'
            },
            class2: {
              id: 'class2',
              name: 'class the second',
              corpus_id: 'corpus1'
            },
            class3: {
              id: 'class3',
              name: 'class the third',
              corpus_id: 'corpus1'
            }
          }
        })
      })

      it('skips if all classes are already loaded', async () => {
        store.state.classification.count.corpus1 = 2
        store.state.classification.hasMLClasses.corpus1 = true
        store.state.classification.classes = {
          class1: {
            id: 'class1',
            name: 'class the first',
            corpus_id: 'corpus1'
          },
          class2: {
            id: 'class2',
            name: 'class the second',
            corpus_id: 'corpus1'
          },
          otherclass: {
            id: 'otherclass',
            name: 'a secret third thing',
            corpus_id: 'corpus2'
          }
        }

        await store.dispatch('classification/listAllCorpusMLClasses', 'corpus1')

        assert.deepStrictEqual(store.history, [
          {
            action: 'classification/listAllCorpusMLClasses',
            payload: 'corpus1'
          }
        ])
      })

      it('paginates if only part of all classes were loaded', async () => {
        store.state.classification.count.corpus1 = 2
        store.state.classification.hasMLClasses.corpus1 = true
        store.state.classification.classes = {
          class1: {
            id: 'class1',
            name: 'class the first',
            corpus_id: 'corpus1'
          },
          otherclass: {
            id: 'otherclass',
            name: 'a secret third thing',
            corpus_id: 'corpus2'
          }
        }

        mock.onGet('/corpus/corpus1/classes/', { params: { page: 1 } }).reply(200, {
          count: 2,
          previous: null,
          next: 'blah blah',
          results: [
            {
              id: 'class2',
              name: 'class the second'
            }
          ]
        })

        await store.dispatch('classification/listAllCorpusMLClasses', 'corpus1')

        assert.deepStrictEqual(store.history, [
          {
            action: 'classification/listAllCorpusMLClasses',
            payload: 'corpus1'
          },
          {
            action: 'classification/listCorpusMLClasses',
            payload: {
              id: 'corpus1',
              page: 1
            }
          },
          {
            mutation: 'classification/setHasMLClasses',
            payload: {
              id: 'corpus1',
              value: true
            }
          },
          {
            mutation: 'classification/setCount',
            payload: {
              id: 'corpus1',
              value: 2
            }
          },
          {
            mutation: 'classification/addMLClass',
            payload: {
              corpus_id: 'corpus1',
              id: 'class2',
              name: 'class the second'
            }
          }
        ])
      })

      it('aborts on the first error', async () => {
        store.state.classification.count.corpus1 = 3
        mock.onGet('/corpus/corpus1/classes/', { params: { page: 1 } }).reply(500)

        await store.dispatch('classification/listAllCorpusMLClasses', 'corpus1')

        assert.deepStrictEqual(store.history, [
          {
            action: 'classification/listAllCorpusMLClasses',
            payload: 'corpus1'
          },
          {
            action: 'classification/listCorpusMLClasses',
            payload: {
              id: 'corpus1',
              page: 1
            }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 500'
            }
          }
        ])
      })

      it('aborts if the backend reports no next page, no matter the count', async () => {
        mock.onGet('/corpus/corpus1/classes/', { params: { page: 1 } }).reply(200, {
          count: 3,
          previous: null,
          next: null,
          results: [
            {
              id: 'class1',
              name: 'class the first'
            }
          ]
        })

        await store.dispatch('classification/listAllCorpusMLClasses', 'corpus1')

        assert.deepStrictEqual(store.history, [
          {
            action: 'classification/listAllCorpusMLClasses',
            payload: 'corpus1'
          },
          {
            action: 'classification/listCorpusMLClasses',
            payload: {
              id: 'corpus1',
              page: 1
            }
          },
          {
            mutation: 'classification/setHasMLClasses',
            payload: {
              id: 'corpus1',
              value: true
            }
          },
          {
            mutation: 'classification/setCount',
            payload: {
              id: 'corpus1',
              value: 3
            }
          },
          {
            mutation: 'classification/addMLClass',
            payload: {
              corpus_id: 'corpus1',
              id: 'class1',
              name: 'class the first'
            }
          }
        ])
      })
    })

    describe('createCorpusMLClass', () => {
      it('creates a new ML class on a corpus', async () => {
        const reply = { id: 'classId', name: 'class name' }
        mock.onPost('/corpus/corpus1/classes/').reply(201, reply)

        await store.dispatch('classification/createCorpusMLClass', { corpusId: 'corpus1', name: 'class name' })

        assert.deepStrictEqual(store.state.classification.classes, { classId: reply })
        assert.deepStrictEqual(store.history, [
          {
            action: 'classification/createCorpusMLClass',
            payload: {
              corpusId: 'corpus1',
              name: 'class name'
            }
          },
          {
            mutation: 'classification/addMLClass',
            payload: reply
          },
          {
            mutation: 'classification/setHasMLClasses',
            payload: { id: 'corpus1', value: true }
          }
        ])
      })

      it('handles errors', async () => {
        mock.onPost('/corpus/corpus1/classes/').reply(400, { detail: 'oh no' })

        await store.dispatch('classification/createCorpusMLClass', { corpusId: 'corpus1', name: 'class name' })

        assert.deepStrictEqual(store.state.classification.classes, {})
        assert.deepStrictEqual(store.history, [
          {
            action: 'classification/createCorpusMLClass',
            payload: {
              corpusId: 'corpus1',
              name: 'class name'
            }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'oh no'
            }
          }
        ])
      })
    })

    describe('editCorpusMLClass', () => {
      it('edits a ML class on a corpus', async () => {
        store.state.classification.classes = { class1: { id: 'class1', name: 'class name' } }
        const reply = { id: 'class1', name: 'new name' }
        mock.onPatch('/corpus/corpus1/classes/class1/').reply(201, reply)

        await store.dispatch('classification/editCorpusMLClass', { corpusId: 'corpus1', classId: 'class1', name: 'new name' })

        assert.deepStrictEqual(store.state.classification.classes, { class1: reply })
        assert.deepStrictEqual(store.history, [
          {
            action: 'classification/editCorpusMLClass',
            payload: {
              corpusId: 'corpus1',
              classId: 'class1',
              name: 'new name'
            }
          },
          {
            mutation: 'classification/addMLClass',
            payload: reply
          }
        ])
      })

      it('handles errors', async () => {
        store.state.classification.classes = { class1: { id: 'class1', name: 'class name' } }
        mock.onPatch('/corpus/corpus1/classes/class1/').reply(400, { detail: 'oh no' })

        await store.dispatch('classification/editCorpusMLClass', { corpusId: 'corpus1', classId: 'class1', name: 'new name' })

        assert.deepStrictEqual(store.state.classification.classes, { class1: { id: 'class1', name: 'class name' } })
        assert.deepStrictEqual(store.history, [
          {
            action: 'classification/editCorpusMLClass',
            payload: {
              corpusId: 'corpus1',
              classId: 'class1',
              name: 'new name'
            }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'oh no'
            }
          }
        ])
      })
    })

    describe('deleteCorpusMLClass', () => {
      it('deletes a ML class from a corpus', async () => {
        store.state.classification.classes = { class1: { id: 'class1', name: 'class name' } }
        mock.onDelete('/corpus/corpus1/classes/class1/').reply(204)

        await store.dispatch('classification/deleteCorpusMLClass', { corpusId: 'corpus1', classId: 'class1' })

        assert.deepStrictEqual(store.state.classification.classes, {})
        assert.deepStrictEqual(store.history, [
          {
            action: 'classification/deleteCorpusMLClass',
            payload: {
              corpusId: 'corpus1',
              classId: 'class1'
            }
          },
          {
            mutation: 'classification/removeMLClass',
            payload: 'class1'
          }
        ])
      })

      it('handles errors', async () => {
        store.state.classification.classes = { class1: { id: 'class1', name: 'class name' } }
        mock.onDelete('/corpus/corpus1/classes/class1/').reply(400, { detail: 'oh no' })

        await store.dispatch('classification/deleteCorpusMLClass', { corpusId: 'corpus1', classId: 'class1' })

        assert.deepStrictEqual(store.state.classification.classes, { class1: { id: 'class1', name: 'class name' } })
        assert.deepStrictEqual(store.history, [
          {
            action: 'classification/deleteCorpusMLClass',
            payload: {
              corpusId: 'corpus1',
              classId: 'class1'
            }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'oh no'
            }
          }
        ])
      })
    })

    describe('validate', () => {
      it('validates a classification', async () => {
        const classification = {
          id: 'classification1',
          state: 'pending'
        }
        const elementId = 'classifiedElement'
        const reply = {
          ...classification,
          state: 'validated'
        }
        mock.onPut('/classifications/classification1/validate/').reply(200, reply)

        const actualReply = await store.dispatch('classification/validate', { id: classification.id, elementId })

        assert.deepStrictEqual(reply, actualReply.data)
        assert.deepStrictEqual(store.history, [
          {
            action: 'classification/validate',
            payload: { id: classification.id, elementId }
          },
          {
            mutation: 'elements/updateClassification',
            payload: {
              value: reply,
              elementId
            }
          }
        ])
      })

      it('handles errors', async () => {
        mock.onPut('/classifications/classification1/validate/').reply(500)

        await store.dispatch('classification/validate', {
          id: 'classification1',
          elementId: 'classifiedElement'
        })

        assert.deepStrictEqual(store.history, [
          {
            action: 'classification/validate',
            payload: {
              id: 'classification1',
              elementId: 'classifiedElement'
            }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 500'
            }
          }
        ])
      })
    })

    describe('reject', () => {
      it('rejects a classification', async () => {
        const classification = {
          id: 'classification1',
          state: 'pending'
        }
        const elementId = 'element1'
        const reply = {
          ...classification,
          state: 'rejected'
        }
        mock.onPut('/classifications/classification1/reject/').reply(200, reply)

        const actualReply = await store.dispatch('classification/reject', { classification, elementId })

        assert.deepStrictEqual(reply, actualReply.data)
        assert.deepStrictEqual(store.history, [
          {
            action: 'classification/reject',
            payload: { classification, elementId }
          },
          {
            mutation: 'elements/updateClassification',
            payload: {
              value: reply,
              elementId
            }
          }
        ])
      })

      it('removes a classification on HTTP 204 (e.g. rejecting a manual classification)', async () => {
        const classification = {
          id: 'classification1',
          state: 'pending'
        }
        const elementId = 'element1'
        mock.onPut('/classifications/classification1/reject/').reply(204)

        await store.dispatch('classification/reject', { classification, elementId })

        assert.deepStrictEqual(store.history, [
          {
            action: 'classification/reject',
            payload: { classification, elementId }
          },
          {
            mutation: 'elements/removeClassification',
            payload: { value: classification, elementId }
          },
          {
            mutation: 'notifications/notify',
            payload: { type: 'info', text: 'Classification deleted due to rejection.' }
          }
        ])
      })

      it('handles errors', async () => {
        mock.onPut('/classifications/classification1/reject/').reply(500)

        await store.dispatch('classification/reject', {
          classification: {
            id: 'classification1'
          },
          elementId: 'classifiedElement'
        })

        assert.deepStrictEqual(store.history, [
          {
            action: 'classification/reject',
            payload: {
              classification: {
                id: 'classification1'
              },
              elementId: 'classifiedElement'
            }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 500'
            }
          }
        ])
      })
    })

    it('creates an element\'s classification', async () => {
      const response = { id: 'classification2', name: 'The best classification' }
      mock.onPost('/classifications/').reply(201, response)
      const payload = {
        mlClass: 'class1',
        elementId: 'element1'
      }
      const classification = await store.dispatch('classification/create', payload)

      assert.deepStrictEqual(classification, response)

      assert.deepStrictEqual(store.history, [
        { action: 'classification/create', payload },
        {
          mutation: 'elements/addClassification',
          payload: { elementId: payload.elementId, value: response }
        }
      ])
      assert.strictEqual(mock.history.post.length, 1)
      assert.deepStrictEqual(mock.history.post[0].data, {
        ml_class: 'class1',
        element: 'element1'
      })
    })
  })
})
