import { assert } from 'chai'
import { setActivePinia } from 'pinia'
import axios from 'axios'
import { useMLResultsStore, useJobsStore, useNotificationStore } from '@/stores'
import { FakeAxios, setUpTestPinia, actionsCompletedPlugin } from '../testhelpers.js'
import { jobsSample, workerVersionsCacheSample, workerVersionsSample } from '../samples.js'

describe('mlresults', () => {
  let pinia, mock, store, notificationStore, jobsStore

  before(() => {
    [pinia] = setUpTestPinia([actionsCompletedPlugin])
    setActivePinia(pinia)
    mock = new FakeAxios(axios)
    store = useMLResultsStore()
    notificationStore = useNotificationStore()
    jobsStore = useJobsStore()
  })

  beforeEach(() => {
    mock.reset()
    store.$reset()
    notificationStore.$reset()
    jobsStore.$reset()
  })

  after('Removing Axios mock', () => {
    mock.restore()
  })

  describe('actions', () => {
    describe('listWorkerVersionsCache', () => {
      it('automatically loads all data', async () => {
        const pagination1 = {
          count: 2,
          previous: null,
          next: '/corpus/corpus_id/versions/?page=2',
          number: 1,
          results: [workerVersionsCacheSample.results[0]]
        }
        const pagination2 = {
          count: 2,
          previous: 'http://previous',
          next: null,
          number: 1,
          results: [workerVersionsCacheSample.results[1]]
        }

        mock.onGet('/corpus/corpus_id/versions/').reply(200, pagination1)
        mock.onGet('/corpus/corpus_id/versions/?page=2').reply(200, pagination2)

        await store.listWorkerVersionsCache('corpus_id')
        // We need to wait for the 2nd async request to be completed
        await store.actionsCompleted()

        assert.deepStrictEqual(store.corpusWorkerVersionsCache, {
          corpus_id: {
            worker_version_cache_1: workerVersionsCacheSample.results[0],
            worker_version_cache_2: workerVersionsCacheSample.results[1]
          }
        })
      })

      it('handles errors', async () => {
        mock.onGet('/corpus/corpus_id/versions/').reply(404, { detail: ['Corpus is private.'] })

        await store.listWorkerVersionsCache('corpus_id')

        assert.deepStrictEqual(store.corpusWorkerVersionsCache, {})

        assert.deepStrictEqual(notificationStore.notifications, [
          {
            id: 0,
            type: 'error',
            text: 'Corpus is private.'
          }
        ])
      })
    })

    describe('deleteWorkerResults', () => {
      it('handles errors', async () => {
        const params = {
          worker_version_id: 'workerversionid',
          model_version_id: 'modelversionid',
          worker_configuration: 'configurationid'
        }
        mock.onDelete('/corpus/corpusid/worker-results/', { params }).reply(418)
        mock.onGet('/jobs/').reply(200, { count: 0 })

        await store.deleteWorkerResults('corpusid', params)
        assert.deepStrictEqual(jobsStore.jobs, {})

        assert.deepStrictEqual(notificationStore.notifications, [
          {
            id: 0,
            type: 'error',
            text: 'Request failed with status code 418'
          }
        ])
      })

      it('deletes ml results', async () => {
        const params = {
          worker_version_id: 'workerversionid',
          model_version_id: 'modelversionid',
          worker_configuration: 'configurationid'
        }
        mock.onDelete('/corpus/corpusid/worker-results/', { params }).reply(204)
        mock.onGet('/jobs/').reply(200, jobsSample)

        const originalState = Object.fromEntries(
          Object.entries(params).map(([key, id]) => [key, { id }])
        )
        store.corpusWorkerVersionsCache = {
          corpusid: { cache_id: { ...originalState } }
        }

        await store.deleteWorkerResults('corpusid', params)
        await jobsStore.actionsCompleted()

        assert.deepStrictEqual(store.corpusWorkerVersionsCache, {
          corpusid: { cache_id: { ...originalState } }
        })
        assert.deepStrictEqual(jobsStore.jobs, {
          jobid: jobsSample[0]
        })

        assert.deepStrictEqual(notificationStore.notifications, [
          {
            id: 0,
            type: 'success',
            text: 'Worker results deletion has been scheduled.'
          }
        ])
      })
    })
  })

  describe('getters', () => {
    describe('corpusWorkerVersions', () => {
      it('lists worker versions related to a corpus', () => {
        store.corpusWorkerVersionsCache = {
          corpus_id: {
            worker_version_cache_1: workerVersionsCacheSample.results[0],
            worker_version_cache_2: workerVersionsCacheSample.results[1],
            worker_version_cache_3: {
              // Same worker version as the 2nd item but without a model version
              ...workerVersionsCacheSample.results[1],
              model_version: null
            }
          }
        }
        assert.deepStrictEqual(store.corpusWorkerVersions('corpus_id'), {
          versionid1: workerVersionsSample.results[0],
          versionid2: workerVersionsSample.results[1]
        })
      })
    })
  })
})
