import axios from 'axios'
import { assert } from 'chai'
import { pick } from 'lodash'
import { createPinia, setActivePinia } from 'pinia'
import { useRightsStore } from '@/stores'
import { assertRejects, FakeAxios } from '../testhelpers'
import { groupSample, membersPageSample } from '../samples'

describe('rights', () => {
  describe('actions', () => {
    let mock, store

    before('Setting up mocks', () => {
      mock = new FakeAxios(axios)
      setActivePinia(createPinia())
      store = useRightsStore()
    })

    afterEach(() => {
      // Remove any handlers, but leave mocking in place
      mock.reset()
      store.$reset()
    })

    after('Removing Axios mock', () => {
      // Remove mocking entirely
      mock.restore()
    })

    describe('createGroup', () => {
      it('creates a new group', async () => {
        mock.onPost('/groups/').reply(201, groupSample)
        const formValues = { name: groupSample.name, public: true }
        await store.createGroup(formValues)

        assert.deepStrictEqual(store.groups, {
          [groupSample.id]: groupSample
        })
      })

      it('throws an error when the data is invalid', async () => {
        mock.onPost('/groups/').reply(400, { name: ['Ensure this field has no more than 64 characters.'] })
        const error = await assertRejects(async () => store.createGroup({ name: groupSample.name }))
        assert.deepStrictEqual(error.response.data, { name: ['Ensure this field has no more than 64 characters.'] })
        assert.deepStrictEqual(store.groups, {})
      })
    })

    describe('updateGroup', () => {
      it('updates an existing group', async () => {
        const reply = {
          id: 'groupid',
          name: 'New name',
          public: false,
          members_count: 1
        }
        mock.onPatch('/group/groupid/').reply(200, reply)
        store.groups = { [groupSample.id]: groupSample }
        const formValues = { name: 'New name', public: false }
        await store.updateGroup('groupid', formValues)

        assert.deepStrictEqual(store.groups, {
          [groupSample.id]: reply
        })
      })
      it('throws an error when the data is invalid', async () => {
        mock.onPatch('/group/groupid/').reply(400, { name: ['Ensure this field has no more than 64 characters.'] })
        store.groups = { [groupSample.id]: groupSample }

        const error = await assertRejects(async () => store.updateGroup(groupSample.id, { name: 'New name' }))

        assert.deepStrictEqual(error.response.data, { name: ['Ensure this field has no more than 64 characters.'] })
        assert.deepStrictEqual(store.groups, { [groupSample.id]: groupSample })
      })
    })

    describe('retrieveGroup', () => {
      it('retrieves group', async () => {
        mock.onGet('/group/groupid/').reply(200, groupSample)
        await store.retrieveGroup('groupid')
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'get',
            url: '/group/groupid/'
          }
        ])
        assert.deepStrictEqual(store.groups, {
          groupid: groupSample
        })
      })
    })

    describe('deleteGroup', () => {
      it('delete an existing group', async () => {
        mock.onDelete('/group/groupid/').reply(204)
        store.groups = { [groupSample.id]: groupSample }

        await store.deleteGroup(groupSample.id)

        assert.deepStrictEqual(store.groups, {})
      })
    })

    describe('listMembers', () => {
      it('get page 1 of members list of a group', async () => {
        mock.onGet('/memberships/', { params: { group: 'groupid', type: 'user', page: 1 } }).reply(200, membersPageSample)

        await store.listMembers({ group: 'groupid', type: 'user', page: 1 })

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'get',
            url: '/memberships/'
          }
        ])
        assert.deepStrictEqual(store.membersPage, membersPageSample)
      })
    })
    describe('listMemberships', () => {
      it('get page 1 of groups list of a user', async () => {
        mock.onGet('/user/memberships/', { params: { page: 1 } }).reply(200, membersPageSample)

        await store.listMemberships({ page: 1 })

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'get',
            url: '/user/memberships/'
          }
        ])
        assert.deepStrictEqual(store.membershipsPage, membersPageSample)
      })
    })
  })
})
